from abc import ABC, abstractmethod
from math import floor
import os

import numpy as np

from model_server.image import GenericImageDataAccessor, InMemoryDataAccessor


class Model(ABC):

    def __init__(self, autoload=True, params=None):
        """
        Abstract base class for an inference model that uses image data as an input.

        :param autoload: automatically load model and dependencies into memory if True
        :param params: Dict[str, str] of arguments e.g. configuration files required to load model
        """
        self.autoload = autoload
        self.params = params
        if self.load():
            self.loaded = True
        else:
            self.loaded = False
            raise CouldNotLoadModelError()
        return None

    @classmethod
    def get_all_subclasses(cls):
        """
        Recursively find all subclasses of Model
        :return: set of all subclasses of Model
        """
        def get_all_subclasses_of(cc):
            return set(cc.__subclasses__()).union(
                [s for c in cc.__subclasses__() for s in get_all_subclasses_of(c)])
        return get_all_subclasses_of(cls)

    @abstractmethod
    def load(self):
        """
        Abstract method that carries out the expectedly time-consuming step of loading a model into memory
        :return: True if successful, else False
        """
        pass

    @abstractmethod
    def infer(self, img: GenericImageDataAccessor) -> (np.ndarray, dict): # return json describing inference result
        if self.autoload:
            self.load()

    def reload(self):
        self.load()


class ImageToImageModel(Model):
    """
    Abstract class for models that receive an image and return an image of the same size
    """

    @abstractmethod
    def infer(self, img) -> (GenericImageDataAccessor, dict):
        super().infer(img)

class DummyImageToImageModel(ImageToImageModel):

    model_id = 'dummy_make_white_square'

    def load(self):
        return True

    def infer(self, img: GenericImageDataAccessor) -> GenericImageDataAccessor:
        super().infer(img)
        w = img.shape_dict['X']
        h = img.shape_dict['Y']
        result = np.zeros([h, w], dtype='uint8')
        result[floor(0.25 * h) : floor(0.75 * h), floor(0.25 * w) : floor(0.75 * w)] = 255
        return InMemoryDataAccessor(data=result)

class Error(Exception):
    pass

class CouldNotLoadModelError(Error):
    pass

class ParameterExpectedError(Error):
    pass